"""Caracara Hosts Module: host online state functions.

In order to avoid the main hosts.py file getting too unwieldly, this file contains
the implementations of the host online state query functions.
"""

# Disable the protected access rule because this file is an extension of the class in hosts.py.
# pylint: disable=protected-access
from __future__ import annotations

from typing import TYPE_CHECKING, List, Union

from caracara.common.batching import batch_get_data
from caracara.common.constants import ONLINE_STATE_DATA_BATCH_SIZE, OnlineState
from caracara.common.exceptions import InvalidOnlineState

if TYPE_CHECKING:
    from caracara.modules.hosts import HostsApiModule


def validate_online_state(
    self: HostsApiModule,
    online_state: Union[OnlineState, str],
):
    """Raise an exception if the online_state_string is not a valid online state.

    Arguments
    ---------
    online_state: OnlineState or str, required
        OnlineState or string to validate.
    """
    if not isinstance(online_state, OnlineState) and online_state not in OnlineState:
        raise InvalidOnlineState(online_state)
    self.logger.debug(f"Validated online state {online_state}")


def get_online_state(
    self: HostsApiModule,
    device_ids: List[str],
) -> List[str]:
    """Return a dictionary containing online state details for every device specified by ID.

    You should only use this endpoint if you already have a list of Device IDs / AIDs,
    and want to retreive data about them directly. If you need to get device data via a
    filter, and you do not yet have a list of Device IDs, you should consider using the
    describe_devices() function.

    Arguments
    ---------
    device_ids: [str], required
        A list of Falcon Device IDs.

    Returns
    -------
    dict: A dictionary containing online state details for every device listed.
    """
    self.logger.info("Obtaining online state data for %s devices", len(device_ids))
    device_online_state_data = batch_get_data(
        device_ids,
        self.hosts_api.get_online_state,
        data_batch_size=ONLINE_STATE_DATA_BATCH_SIZE,
    )
    return device_online_state_data


def filter_device_ids_by_online_state(
    self: HostsApiModule,
    device_ids: List[str],
    online_state: Union[OnlineState, str],
) -> List[str]:
    """Filter a list of device IDs by an online state.

    Arguments
    ---------
    device_ids: List[str]
        Device IDs to filter against
    online_state: OnlineState or str
        Online state to filter device IDs on. Options are "online", "offline", "unknown"

    Returns
    -------
    List[str]: A list of device IDs with the specified online state
    """
    self.validate_online_state(online_state)

    device_state_data = self.get_online_state(device_ids)

    return list(
        filter(
            lambda device_id: device_state_data[device_id]["state"] == online_state,
            device_state_data,
        )
    )



"""Falcon Sensor Download API."""

from falconpy import OAuth2, SensorDownload

from caracara.common.exceptions import GenericAPIError
from caracara.common.module import FalconApiModule, ModuleMapper


class SensorDownloadApiModule(FalconApiModule):
    """
    Sensor Download API Module.

    Whilst this module will eventually contain the logic required to inspect the available
    versions of the Falcon sensor available for download, its primary purpose is to allow
    other modules to retrieve the (C)CID.
    """

    name = "CrowdStrike Sensor Download API Module"
    help = "Download the Falcon sensor installer and fetch the installation CCID"

    def __init__(self, api_authentication: OAuth2, mapper: ModuleMapper):
        """Construct an instance of the SensorDownloadApiModule class."""
        super().__init__(api_authentication, mapper)
        self.logger.debug("Configuring the FalconPy Sensor Download API")
        self.sensor_download_api = SensorDownload(auth_object=self.api_authentication)

    def get_cid(self, include_checksum: bool = False) -> str:
        """Obtain the Customer ID (CID) associated with the currently authenticated API token.

        If include_checksum=True, Checksummed CID (CCID) will be returned, which includes a hyphen
        and a two character checksum appended to the CID. The CCID is required for installing the
        sensor.
        If include_checksum=False, the resultant CID will be lower-cased for broad compatibility
        outside of sensor installers.
        """
        self.logger.info("Obtaining the CCID from the cloud using the Sensor Download API")

        response = self.sensor_download_api.get_sensor_installer_ccid()
        self.logger.debug(response)

        try:
            ccid = response["body"]["resources"][0]
        except (KeyError, IndexError) as exc:
            self.logger.info(
                "Failed to retrieve the CCID from the cloud. Check your API credentials."
            )
            raise GenericAPIError(response["body"]["errors"]) from exc

        if include_checksum:
            return ccid

        # Strip the hyphen and two character checksum if we just need the CID, then send to
        # lower case.
        cid = ccid[:-3].lower()
        return cid



"""Caracara Sensor Policies API Module."""

__all__ = [
    "SensorUpdatePoliciesApiModule",
]

from caracara.modules.sensor_update_policies.sensor_update_policies import (
    SensorUpdatePoliciesApiModule,
)



"""Sensor Update Policy APIs.

This module allows for interaction with the Sensor Update policies, and facilitates
the retrieval of maintenance tokens.
"""

import datetime
from typing import Optional

from falconpy import OAuth2, SensorUpdatePolicies

from caracara.common.exceptions import BaseCaracaraError
from caracara.common.module import FalconApiModule, ModuleMapper


class SensorUpdatePoliciesApiModule(FalconApiModule):
    """This module facilitates interactions with the Sensor Update Policies API."""

    name = "CrowdStrike Sensor Update Policies API Module"
    help = "Interact with the Sensor Update Policies API, and get maintenance tokens"

    def __init__(self, api_authentication: OAuth2, mapper: ModuleMapper):
        """Construct an instance of the SensorUpdatePoliciesApiModule."""
        super().__init__(api_authentication, mapper)

        self.logger.debug("Configuring the FalconPy Sensor Update Policies module")
        self.sensor_update_policies_api = SensorUpdatePolicies(auth_object=self.api_authentication)

    def get_maintenance_token(
        self,
        device_id: str,
        audit_message: Optional[str] = None,
    ) -> str:
        """Get the maintenance token for a device.

        Arguments
        ---------
        device_id: str, required
            A Falcon Device ID (AID) to get the maintenance token for.

        audit_message: str or None
            A text string containing an audit message for the retrieval of the maintenance token.
            If left blank / None, Caracara will generate one for you.

        Returns
        -------
        str: The maintenance token.
        """
        if audit_message is None:
            timestamp = datetime.datetime.now(
                tz=datetime.timezone.utc,
            ).strftime("%Y-%m-%d %H:%M:%S")

            audit_message = f"Generated via Caracara at {timestamp} UTC"

        response = self.sensor_update_policies_api.reveal_uninstall_token(
            audit_message=audit_message,
            device_id=device_id,
        )

        if response["status_code"] == 200:
            body = response["body"]
            return body["resources"][0]["uninstall_token"]

        raise BaseCaracaraError(
            "The API operation failed to generate a maintenance token."
            "Check that you have the Sensor Update Policies - Write permission "
            "enabled on your API token."
        )

    def get_bulk_maintenance_token(self, audit_message: Optional[str] = None) -> str:
        """Get the bulk maintenance token for a Falcon tenant.

        Arguments
        ---------
        audit_message: str or None
            A text string containing an audit message for the retrieval of the maintenance token.
            If left blank / None, Caracara will generate one for you.

        Returns
        -------
        str: The bulk maintenance token.
        """
        return self.get_maintenance_token(device_id="MAINTENANCE", audit_message=audit_message)



"""Falcon Users API."""

from functools import partial
from typing import Dict, List, Optional, Union

from falconpy import OAuth2, UserManagement

from caracara.common.batching import batch_get_data
from caracara.common.exceptions import GenericAPIError
from caracara.common.module import FalconApiModule, ModuleMapper
from caracara.common.pagination import (
    all_pages_numbered_offset_parallel,
    generic_parallel_list_execution,
)
from caracara.filters import FalconFilter
from caracara.filters.decorators import filter_string


class UsersApiModule(FalconApiModule):
    """
    Users API module.

    This module provides the logic to interact with the Falcon User Management
    APIs. With the functions provided herein, one can list and create users, and
    assign access roles.
    """

    name = "CrowdStrike User Management API Module"
    help = "Describe, create, delete and edit users in a Falcon tenant"

    def __init__(self, api_authentication: OAuth2, mapper: ModuleMapper):
        """Construct an instance of the UsersApiModule class."""
        super().__init__(api_authentication, mapper)
        self.logger.debug("Configuring the FalconPy User Management API")
        self.user_management_api = UserManagement(auth_object=self.api_authentication)

    @filter_string
    def get_user_uuids(self, filters: Optional[Union[FalconFilter, str]] = None) -> List[str]:
        """Get a list of IDs of users (UUIDs) configured in the Falcon tenant."""
        self.logger.info("Obtaining a list of all users in the Falcon tenant")

        query_users_func = partial(self.user_management_api.query_users, filter=filters)
        user_uuids: List[str] = all_pages_numbered_offset_parallel(
            query_users_func,
            self.logger,
            500,
        )
        return user_uuids

    def get_user_data(self, user_uuids: List[str]) -> Dict[str, Dict]:
        """Fetch a dictionary of data for each of the User IDs (UUIDs) passed into the function."""
        self.logger.info("Obtaining data for the %d User IDs provided", len(user_uuids))

        user_data: Dict[str, Dict] = batch_get_data(
            user_uuids,
            self.user_management_api.retrieve_users,
        )
        return user_data

    @filter_string
    def describe_users(
        self,
        filters: Optional[Union[FalconFilter, str]] = None,
        user_uuids: Optional[List[str]] = None,
    ) -> Dict[str, Dict]:
        """Describe the users in a Falcon tenant.

        If a list of UUIDs is provided, the filters will be ignored and only users with those UUIDs
        will be described.
        """
        self.logger.info("Describing users")

        if not user_uuids:
            user_uuids = self.get_user_uuids(filters=filters)

        user_data = self.get_user_data(user_uuids)
        user_roles = self.get_assigned_user_roles(user_uuids)

        # Set up an empty set to contain all roles assigned to each user
        for user_data_dict in user_data.values():
            user_data_dict["roles"] = set()

        # Iterate over all roles in the CID and add them to each user's role set
        for user_role in user_roles:
            if user_role["uuid"] in user_data:
                user_data[user_role["uuid"]]["roles"].add(user_role["role_id"])

        # Convert the sets to sorted list objects so that they can be JSON serialised
        for user_data_dict in user_data.values():
            user_data_dict["roles"] = sorted(list(user_data_dict["roles"]))

        return user_data

    def get_available_role_ids(self) -> List[str]:
        """Obtain a list of role IDs enabled on the Falcon tenant."""
        self.logger.info("Fetching a list of role IDs")

        # Endpoint does not support offsets, so we do not need to parallelise here
        role_ids: List[str] = self.user_management_api.query_roles()["body"]["resources"]
        return role_ids

    def get_role_information(self, role_ids: List[str]) -> Dict[str, Dict]:
        """Fetch a dictionary describing each of the Role IDs passed into the function."""
        self.logger.info("Getting information on the %d role IDs provided", len(role_ids))

        role_info: Dict[str, Dict] = batch_get_data(
            role_ids,
            self.user_management_api.get_roles_mssp,
        )
        return role_info

    def describe_available_roles(self) -> Dict[str, Dict]:
        """Describe the roles that are available within the current Falcon tenant."""
        self.logger.info("Describing available roles")

        role_ids = self.get_available_role_ids()
        role_info = self.get_role_information(role_ids)
        return role_info

    def get_assigned_user_roles(self, user_uuids: List[str]) -> Dict[str, Dict]:
        """Retrieve a list of roles assigned to a list of User UUIDs."""
        self.logger.info("Retrieving roles for %d User IDs", len(user_uuids))

        get_user_grants_func = partial(
            self.user_management_api.get_user_grants,
            direct_only=False,
            sort="cid",
        )
        user_roles = generic_parallel_list_execution(
            get_user_grants_func,
            logger=self.logger,
            param_name="user_uuid",
            value_list=user_uuids,
        )
        return user_roles

    def add_user_roles(self, user_uuid: str, role_ids: List[str]):
        """Add Roles via role_id to a User via UUID. Sensor Download READ permissions required."""
        self.logger.info("Granting roles %s to user %s", role_ids, user_uuid)

        cid = self.mapper.sensor_download.get_cid()

        add_roles_response = self.user_management_api.user_roles_action(
            action="grant", cid=cid, role_ids=role_ids, uuid=user_uuid
        )
        self.logger.debug(add_roles_response)

        return add_roles_response["status_code"] == 200

    def add_user(self, first_name: str, last_name: str, email_address: str) -> Dict[str, str]:
        """Add a user via first name, last name, and email address."""
        self.logger.info("Creating a new user, %s", email_address)
        new_user_response = self.user_management_api.create_user_mssp(
            first_name=first_name, last_name=last_name, uid=email_address
        )
        if new_user_response["body"]["resources"] is None:
            raise GenericAPIError(new_user_response["body"]["errors"])

        new_user_data = new_user_response["body"]["resources"][0]
        return new_user_data

    def delete_user(self, uuid) -> bool:
        """DESTRUCTIVE ACTION: Permanently delete a user via UUID.

        If the deletion succeeded, this function will return True.
        Otherwise, it will return False.
        """
        self.logger.warning("Permanently deleting user with UUID %s", uuid)
        deleted_user_response = self.user_management_api.delete_user_mssp(self, user_uuid=uuid)
        return deleted_user_response["status_code"] == 200

    def get_uuid_by_email(self, email: str) -> str:
        """Get UUID based on user email address."""
        self.logger.info("Getting UUID")
        uuid_response = self.user_management_api.retrieve_user_uuid(uid=email)
        try:
            uuid = uuid_response["body"]["resources"][0]
        except (KeyError, IndexError) as exc:
            raise GenericAPIError(uuid_response["body"]["errors"]) from exc

        return uuid



"""Caracara: CrowdStrike Style Radio Dialog for Prompt Toolkit."""

from __future__ import annotations

from typing import Any, Sequence, TypeVar

from prompt_toolkit.application import Application
from prompt_toolkit.application.current import get_app
from prompt_toolkit.formatted_text import AnyFormattedText
from prompt_toolkit.key_binding.bindings.focus import focus_next, focus_previous
from prompt_toolkit.key_binding.defaults import load_key_bindings
from prompt_toolkit.key_binding.key_bindings import KeyBindings, merge_key_bindings
from prompt_toolkit.layout import Layout
from prompt_toolkit.layout.containers import AnyContainer, HSplit
from prompt_toolkit.styles import BaseStyle, Style, merge_styles
from prompt_toolkit.widgets import Button, Dialog, Label, RadioList

# Prompt Toolkit uses _T for the generic type
_T = TypeVar("_T")


def _create_app(dialog: AnyContainer, style: BaseStyle | None) -> Application[Any]:
    """Turn a container into a Prompt Toolkit Application."""
    bindings = KeyBindings()
    bindings.add("tab")(focus_next)
    bindings.add("s-tab")(focus_previous)

    return Application(
        layout=Layout(dialog),
        key_bindings=merge_key_bindings([load_key_bindings(), bindings]),
        mouse_support=True,
        style=style,
        full_screen=True,
    )


def _return_none() -> None:
    """Button handler that returns None.

    This is used by cancel buttons.
    """
    get_app().exit()


CS_STYLE = Style.from_dict(
    {
        "button": "fg:#1A1A1A bold",
        "button.arrow": "fg:#FC0000 bold",
        "button.focused": "fg:#FFFFFF bg:#2F8BAA",
        "button.text": "fg:#1A1A1A bold",
        "dialog": "bg:#58595B",
        "dialog.body": "bg:#F3F3F4",
        "dialog.body label": "fg:#FC0000",
        "dialog frame.label": "bg:#F3F4F4 #1A1A1A",
        "dialog shadow": "bg:#68696B",
        "frame.label": "fg:#FC0000",
        "label": "#FC0000",
        "radio-list": "#1A1A1A",
        "radio-checked": "fg:#2F8BAA",
    }
)


def csradiolist_dialog(  # pylint: disable=too-many-arguments
    title: AnyFormattedText = "",
    text: AnyFormattedText = "",
    ok_text: str = "Ok",
    cancel_text: str = "Cancel",
    values: Sequence[tuple[_T, AnyFormattedText]] | None = None,
    default: _T | None = None,
    style: BaseStyle | None = None,
) -> Application[_T]:
    """Display a CrowdStrike styled simple list of elements that the user can choose from.

    This function is heavily based on the radiolist_dialog function provided by Prompt Toolkit,
    with some tweaks in place to handle a lack of cancel button (where desired) and a default
    set of styles.

    Only one element can be selected at a time using Arrow keys and Enter.
    The focus can be moved between the list and the Ok/Cancel button with tab.
    """
    if values is None:
        values = []

    def ok_handler() -> None:
        get_app().exit(result=radio_list.current_value)

    radio_list = RadioList(values=values, default=default)

    buttons = [Button(text=ok_text, handler=ok_handler)]

    if cancel_text is not None:
        buttons.append(Button(text=cancel_text, handler=_return_none))

    dialog = Dialog(
        title=title,
        body=HSplit(
            [Label(text=text, dont_extend_height=True), radio_list],
            padding=1,
        ),
        buttons=buttons,
        with_background=True,
    )

    if style is None:
        style = CS_STYLE
    else:
        style = merge_styles([CS_STYLE, style])

    app = _create_app(dialog, style)

    return app



r"""Caracara base API interface client.

CrowdStrike Caracara
    _______ __ __             __
   |   _   |  |__.-----.-----|  |_
   |.  1___|  |  |  -__|     |   _|
   |.  |___|__|__|_____|__|__|____|
   |:  1   |
   |::.. . |     for FalconPy
   `-------'
"""

import logging

try:
    from falconpy import OAuth2, confirm_base_region, confirm_base_url
except ImportError as no_falconpy:
    raise SystemExit("The crowdstrike-falconpy library is not installed.") from no_falconpy

from caracara_filters import FQLGenerator

from caracara.common.interpolation import VariableInterpolator
from caracara.common.meta import user_agent_string
from caracara.common.module import ModuleMapper
from caracara.modules import (
    CustomIoaApiModule,
    FlightControlApiModule,
    HostsApiModule,
    PreventionPoliciesApiModule,
    ResponsePoliciesApiModule,
    RTRApiModule,
    SensorDownloadApiModule,
    SensorUpdatePoliciesApiModule,
    UsersApiModule,
)


class Client:
    """Falcon API client.

    This class exposes all available Falcon API commands, and proxies requests through
    the FalconPy library.
    """

    class FalconFilter(FQLGenerator):
        """Falcon Filter wrapper class.

        FalconFilter has now been replaced with the externally provided FQLGenerator functionality.
        This subclass allows pre-existing code to continue referencing FalconFilter.
        """

    def __init__(  # pylint: disable=R0913,R0914,R0915
        self,
        client_id: str = None,
        client_secret: str = None,
        cloud_name: str = "auto",
        member_cid: str = None,
        ssl_verify: bool = True,
        timeout: float = None,
        proxy: str = None,
        user_agent: str = None,
        verbose: bool = False,
        falconpy_authobject: OAuth2 = None,
    ):
        """Configure a Caracara Falcon API Client object."""
        self.logger = logging.getLogger(__name__)

        if client_id is None and client_secret is None and falconpy_authobject is None:
            raise ValueError(
                "You must provide either a Client ID and Client Secret, "
                "or a pre-created FalconPy OAuth2 object"
            )

        if client_id is not None and client_secret is None:
            raise ValueError("You cannot provide a Client ID without a Client Secret")

        if client_id is not None and falconpy_authobject is not None:
            raise ValueError(
                "Please provide either a Client ID/Client Secret pair, "
                "or a pre-created FalconPy OAuth2 object, but not both"
            )

        self.logger.info("Setting up the Caracara client and configuring authentication")

        if client_id:
            # Load all parameters for the FalconPy authentication object into a dictionary
            # and handle environment variable interpolation
            interpolator = VariableInterpolator()
            cloud_name = interpolator.interpolate(cloud_name)
            client_id = interpolator.interpolate(client_id)
            client_secret = interpolator.interpolate(client_secret)
            member_cid = interpolator.interpolate(member_cid)
            user_agent = interpolator.interpolate(user_agent)
            proxy = interpolator.interpolate(proxy)
            user_agent = interpolator.interpolate(user_agent)

            if user_agent:
                user_agent = f"{user_agent} ({user_agent_string()})"
            else:
                user_agent = user_agent_string()

            self.logger.debug("User agent: %s", user_agent)

            auth_keys = {
                "base_url": cloud_name,
                "client_id": client_id,
                "client_secret": client_secret,
                "member_cid": member_cid,
                "proxy": proxy,
                "user_agent": user_agent,
                "ssl_verify": ssl_verify,
                "timeout": timeout,
            }

            self.logger.info(
                "Client ID: %s; Cloud: %s; Member CID: %s",
                client_id,
                cloud_name,
                member_cid,
            )
            self.logger.debug("SSL verification is %s", ssl_verify)
            self.logger.debug("Timeout: %s", str(timeout))
            self.logger.debug("Configured proxy: %s", proxy)

            # Remove all None values, as we do not wish to override any FalconPy defaults
            for k in list(auth_keys.keys()):
                if auth_keys[k] is None:
                    del auth_keys[k]

            self.verbose = verbose
            self.logger.debug("Verbose mode: %s", verbose)

            base_url = confirm_base_region(confirm_base_url(provided_base=cloud_name))
            self.logger.info("Base URL: %s", base_url)

            self.logger.debug("Configuring api_authentication object as an OAuth2 object")
            self.api_authentication = OAuth2(**auth_keys)
        elif falconpy_authobject:
            self.logger.info(
                "Using pre-created FalconPy OAuth2 object. All other options will be ignored"
            )
            self.api_authentication = falconpy_authobject
        else:
            raise TypeError("Impossible authentication scenario")

        self.logger.info("Requesting API token")
        self.api_authentication.token()  # Need to force the authentication to resolve the base_url
        self.logger.info("Resolved Base URL: %s", self.api_authentication.base_url)

        mapper = ModuleMapper()

        # Configure modules here so that IDEs can pick them up
        self.logger.debug("Setting up Custom IOA module")
        self.custom_ioas = CustomIoaApiModule(self.api_authentication, mapper)
        mapper.custom_ioas = self.custom_ioas

        self.logger.debug("Setting up the Flight Control module")
        self.flight_control = FlightControlApiModule(self.api_authentication, mapper)
        mapper.flight_control = self.flight_control

        self.logger.debug("Setting up the Hosts module")
        self.hosts = HostsApiModule(self.api_authentication, mapper)
        mapper.hosts = self.hosts

        self.logger.debug("Setting up the Prevention Policies module")
        self.prevention_policies = PreventionPoliciesApiModule(self.api_authentication, mapper)
        mapper.prevention_policies = self.prevention_policies

        self.logger.debug("Setting up the Response Policies module")
        self.response_policies = ResponsePoliciesApiModule(self.api_authentication, mapper)
        mapper.response_policies = self.response_policies

        self.logger.debug("Setting up the RTR module")
        self.rtr = RTRApiModule(self.api_authentication, mapper)
        mapper.rtr = self.rtr

        self.logger.debug("Setting up the Sensor Download module")
        self.sensor_download = SensorDownloadApiModule(self.api_authentication, mapper)
        mapper.sensor_download = self.sensor_download

        self.logger.debug("Setting up the Sensor Update Policies module")
        self.sensor_update_policies = SensorUpdatePoliciesApiModule(self.api_authentication, mapper)
        mapper.sensor_update_policies = self.sensor_update_policies

        self.logger.debug("Setting up the Users module")
        self.users = UsersApiModule(self.api_authentication, mapper)
        mapper.users = self.users

        self.logger.info("Caracara client configured")

    def __enter__(self):
        """Allow for entry as a context manager."""
        self.logger.debug("Entering Caracara context manager")
        return self

    def __exit__(self, *args):
        """Discard our token when we exit the context."""
        self.logger.info("Discarding API token on Client exit")
        self.api_authentication.revoke(self.api_authentication.token_value)
        self.logger.debug("Discarding Caracara context manager")
        return args



"""Generic Caracara API module.

This module contains the the FalconApiModule class, which represents a generic
Caracara API module. All modules, including Hosts, Prevention Policies, etc.
derive from this abstract base class.
"""

import logging
from abc import ABC, abstractmethod

from falconpy import OAuth2


# pylint: disable=R0903
class ModuleMapper:
    """Empty container class to allow modules to map into other modules.

    This is deliberately empty to start with as the modules are loaded into
    the class dynamically when the Client is set up. This allows modules to
    call into one another post-initialisation.
    """


class FalconApiModule(ABC):
    """
    Meta class for a generic Caracara API Module.

    Each module provides API Methods.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """Store the name for the developer to use when instantiating the API module."""

    @property
    @abstractmethod
    def help(self) -> str:
        """Store the help string to be made available for each API module."""

    def __init__(self, api_authentication: OAuth2, mapper: ModuleMapper):
        """Configure a Caracara API module with a FalconPy OAuth2 module."""
        class_name = self.__class__.__name__
        self.logger = logging.getLogger(f"caracara.modules.{class_name}")
        self.logger.debug("Initialising API module: %s", class_name)

        self.api_authentication = api_authentication
        self.mapper = mapper



"""
Caracara Modules Package Initialisation.

Proides pipework to link together the various modules within Caracara and expose them to
the Client object at setup.
"""

__all__ = [
    "CustomIoaApiModule",
    "FlightControlApiModule",
    "HostsApiModule",
    "PreventionPoliciesApiModule",
    "ResponsePoliciesApiModule",
    "RTRApiModule",
    "SensorDownloadApiModule",
    "SensorUpdatePoliciesApiModule",
    "UsersApiModule",
]

from caracara.modules.custom_ioa import CustomIoaApiModule
from caracara.modules.flight_control import FlightControlApiModule
from caracara.modules.hosts import HostsApiModule
from caracara.modules.prevention_policies import PreventionPoliciesApiModule
from caracara.modules.response_policies import ResponsePoliciesApiModule
from caracara.modules.rtr import RTRApiModule
from caracara.modules.sensor_download import SensorDownloadApiModule
from caracara.modules.sensor_update_policies import SensorUpdatePoliciesApiModule
from caracara.modules.users import UsersApiModule



"""Caracara Custom Indicator of Attack (IOA) module."""

__all__ = [
    "CustomIoaApiModule",
    "CustomIoaRule",
    "IoaRuleGroup",
]

from caracara.modules.custom_ioa.custom_ioa import CustomIoaApiModule
from caracara.modules.custom_ioa.rules import CustomIoaRule, IoaRuleGroup



r"""Hosts and Host Group APIs.

 ___ ___             __          ___ ___                                                    __
|   Y   .-----.-----|  |_.-----.|   Y   .---.-.-----.---.-.-----.-----.--------.-----.-----|  |_
|.  1   |  _  |__ --|   _|__ --||.      |  _  |     |  _  |  _  |  -__|        |  -__|     |   _|
|.  _   |_____|_____|____|_____||. \_/  |___._|__|__|___._|___  |_____|__|__|__|_____|__|__|____|
|:  |   |                       |:  |   |                 |_____|
|::.|:. |                       |::.|:. |                                   CrowdStrike Caracara
`--- ---'                       `--- ---'

This module handles interactions with the CrowdStrike Falcon Hosts and Host Group APIs.
"""

from functools import partial
from typing import Dict, List, Optional, Union

from falconpy import HostGroup, Hosts, OAuth2

from caracara.common.batching import batch_get_data
from caracara.common.constants import OnlineState
from caracara.common.exceptions import GenericAPIError
from caracara.common.module import FalconApiModule, ModuleMapper
from caracara.common.pagination import all_pages_token_offset
from caracara.filters import FalconFilter
from caracara.filters.decorators import filter_string


# Allow importing from other files into this class so that the file does not end up
# being over 1000 lines long.
# pylint: disable=import-outside-toplevel
class HostsApiModule(FalconApiModule):
    """The HostsApiModule represents interactions with the Hosts and Host Group APIs."""

    name = "CrowdStrike Hosts API Module"
    help = "Interact with hosts and host groups within your Falcon tenant."

    def __init__(self, api_authentication: OAuth2, mapper: ModuleMapper):
        """Construct an instance of HostApiModule class."""
        super().__init__(api_authentication, mapper)

        self.logger.debug("Configuring the FalconPy Hosts API")
        self.hosts_api = Hosts(auth_object=self.api_authentication)

        self.logger.debug("Configuring the FalconPy Host Group API")
        self.host_group_api = HostGroup(auth_object=self.api_authentication)

    # Import containment functions
    from caracara.modules.hosts._containment import contain, release

    # Import data history functions
    from caracara.modules.hosts._data_history import (
        describe_login_history,
        describe_network_address_history,
        describe_state,
    )

    # Import functions pertaining to host groups
    from caracara.modules.hosts._groups import (
        _create_host_group,
        _perform_group_action,
        add_to_group,
        create_group,
        delete_group,
        describe_group_member_ids,
        describe_group_members,
        describe_groups,
        get_group_ids,
        get_group_member_ids,
        get_group_members,
        group,
        remove_from_group,
        ungroup,
        update_group,
    )

    # Import functions to hide, unhide and describe hidden devices
    from caracara.modules.hosts._hiding import (
        describe_hidden_devices,
        get_hidden_ids,
        hide,
        unhide,
    )

    # Import functions to filter by device online state
    from caracara.modules.hosts._online_state import (
        filter_device_ids_by_online_state,
        get_online_state,
        validate_online_state,
    )

    # Import functions to create and modify device tags
    from caracara.modules.hosts._tagging import (
        _create_tag_list,
        _update_device_tags,
        tag,
        untag,
    )

    # Static methods have to be set within the class
    _create_tag_list = staticmethod(_create_tag_list)

    def _perform_action(
        self,
        action_name: str,
        device_ids: List[str],
    ) -> Dict:
        """Perform the specified action against the list of targets."""
        returned = self.hosts_api.perform_action(ids=device_ids, action_name=action_name)["body"]

        if returned["errors"]:
            raise GenericAPIError(returned["errors"])

        return returned

    @filter_string
    def describe_devices(
        self,
        filters: Union[FalconFilter, str] = None,
        online_state: Optional[Union[OnlineState, str]] = None,
        enrich_with_online_state: Optional[bool] = False,
    ) -> Dict[str, Dict]:
        """Return a dictionary containing details for every device matching the provided filter.

        Arguments
        ---------
        filters: Union[FalconFilter, str], optional
            Filters to apply to the device search.
        online_state: OnlineState or str, optional
            Device online state to filter devices on. Options are "online", "offline", "unknown"

        Returns
        -------
        dict: A dictionary containing details for every device discovered.
        """
        self.logger.info("Describing devices according to the filter string %s", filters)
        device_ids = self.get_device_ids(filters)

        if enrich_with_online_state or online_state is not None:
            # Collect state data
            device_state_data = self.get_online_state(device_ids)

            # Filter by online state, if applicable.
            if online_state is not None:
                self.validate_online_state(online_state)
                device_ids = list(
                    filter(
                        lambda key: device_state_data[key]["state"] == online_state,
                        device_state_data,
                    )
                )

        device_data = self.get_device_data(device_ids)

        if enrich_with_online_state:
            # Enrich the results with the online state field
            for device_id, data in device_data.items():
                data["state"] = device_state_data[device_id]["state"]

        return device_data

    def get_device_data(
        self,
        device_ids: List[str],
    ) -> Dict[str, Dict]:
        """Return a dictionary containing details for every device specified by ID.

        You should only use this endpoint if you already have a list of Device IDs / AIDs,
        and want to retreive data about them directly. If you need to get device data via a
        filter, and you do not yet have a list of Device IDs, you should consider using the
        describe_devices() function.

        Arguments
        ---------
        device_ids: [str], required
            A list of Falcon Device IDs.

        Returns
        -------
        dict: A dictionary containing details for every device listed.
        """
        self.logger.info("Obtaining data for %s devices", len(device_ids))
        device_data = batch_get_data(device_ids, self.hosts_api.get_device_details)
        return device_data

    @filter_string
    def get_device_ids(
        self,
        filters: Union[FalconFilter, str] = None,
        online_state: Optional[Union[OnlineState, str]] = None,
    ) -> List[str]:
        """Return a list of IDs (string) for every device in your Falcon tenant.

        Arguments
        ---------
        filters: Union[FalconFilter, str], optional
            Filters to apply to the device search.
        online_state: OnlineState or str, optional
            Device online state to filter devices on. Options are "online", "offline", "unknown"

        Returns
        -------
        List[str]: A list of all device IDs discovered.
        """
        self.logger.info("Searching for device IDs using the filter string %s", filters)
        func = partial(self.hosts_api.query_devices_by_filter_scroll, filter=filters)
        id_list: List[str] = all_pages_token_offset(func=func, logger=self.logger)

        # Filter by online state, if applicable.
        if online_state is not None:
            id_list = self.filter_device_ids_by_online_state(id_list, online_state=online_state)

        return id_list



"""Caracara: Common functions and data imports."""

__all__ = [
    "DEFAULT_DATA_BATCH_SIZE",
    "FalconApiModule",
    "Policy",
    "SCROLL_BATCH_SIZE",
    "user_agent_string",
]

from caracara.common.constants import DEFAULT_DATA_BATCH_SIZE, SCROLL_BATCH_SIZE
from caracara.common.meta import user_agent_string
from caracara.common.module import FalconApiModule
from caracara.common.policy_wrapper import Policy



"""Common constants to be shared throughout Caracara."""

from enum import Enum, EnumMeta

# Batch size of data downloaded via a multi-threaded data pull
DEFAULT_DATA_BATCH_SIZE = 500

# Batch size of data downloaded via a multi-threaded data pull from the online state endpoint
ONLINE_STATE_DATA_BATCH_SIZE = 100

# Default pagination limit
PAGINATION_LIMIT = 100

# Batch size of data downloaded from scrolled endpoints
SCROLL_BATCH_SIZE = 5000

# Host group batch size
HOST_GROUP_SCROLL_BATCH_SIZE = 100

# List of all platforms supported by Falcon
PLATFORMS = [
    "Linux",
    "Mac",
    "Windows",
]

#
DEFAULT_COMMENT = "This action was performed by the Caracara Python library."


# Device online states classes
class MetaEnum(EnumMeta):
    """Overrided class for the use of the in operator and to query for all valid enum values."""

    def __init__(cls, *kwargs):
        """Store all possible values of the Enum subclass."""
        cls.VALUES = [state.value for state in cls.__members__.values()]
        super().__init__(kwargs)

    def __contains__(cls: Enum, item: str) -> bool:
        """Override the __contains__ method to use the in operator with Enum subclasses."""
        return item in cls.VALUES


class OnlineState(Enum, metaclass=MetaEnum):
    """
    Falcon OnlineState class.

    Enum class of valid device online states. Valid states are 'online', 'offline', and 'unknown'.
    """

    ONLINE = "online"
    OFFLINE = "offline"
    UNKNOWN = "unknown"

    def __eq__(self, item: str) -> bool:
        """Override the __eq__ method for strings to compare the OnlineState's string value."""
        return str(self.value) == item



"""Caracara Real Time Response (RTR) module."""

__all__ = [
    "BatchGetCmdRequest",
    "GetFile",
    "InnerRTRBatchSession",
    "RTRApiModule",
    "RTRBatchSession",
]

from caracara.modules.rtr.batch_session import (
    BatchGetCmdRequest,
    InnerRTRBatchSession,
    RTRBatchSession,
)
from caracara.modules.rtr.get_file import GetFile
from caracara.modules.rtr.rtr import RTRApiModule



"""Caracara PreventionPoliciesApiModule."""

__all__ = [
    "PreventionPoliciesApiModule",
]

from caracara.modules.prevention_policies.prevention_policies import (
    PreventionPoliciesApiModule,
)
